@@ -31,31 +31,12 @@
 
 G_BEGIN_DECLS
 
-#define GDK_PARENT_RELATIVE_BG ((GdkPixmap *)1L)
-#define GDK_NO_BG ((GdkPixmap *)2L)
-
-#define GDK_WINDOW_TYPE(d) (((GdkWindowObject*)(GDK_WINDOW (d)))->window_type)
-#define GDK_WINDOW_DESTROYED(d) (((GdkWindowObject*)(GDK_WINDOW (d)))->destroyed)
-
 void gdk_window_destroy_notify	     (GdkWindow *window);
 
 void gdk_synthesize_window_state (GdkWindow     *window,
                                   GdkWindowState unset_flags,
                                   GdkWindowState set_flags);
 
-gboolean _gdk_window_event_parent_of (GdkWindow *parent,
-                                      GdkWindow *child);
-
-/* Tests whether a pair of x,y may cause overflows when converted to Pango
- * units (multiplied by PANGO_SCALE).  We don't allow the entire range, leave
- * some space for additions afterwards, to be safe...
- */
-#define GDK_PANGO_UNITS_OVERFLOWS(x,y) (G_UNLIKELY ( \
-	(y) >= PANGO_PIXELS (G_MAXINT-PANGO_SCALE)/2 || \
-	(x) >= PANGO_PIXELS (G_MAXINT-PANGO_SCALE)/2 || \
-	(y) <=-PANGO_PIXELS (G_MAXINT-PANGO_SCALE)/2 || \
-	(x) <=-PANGO_PIXELS (G_MAXINT-PANGO_SCALE)/2))
-
 G_END_DECLS
 
 #endif /* __GDK_PRIVATE_H__ */
